#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan 14 15:24:08 2019

@author: tladd

Calculate evolution of the conformal mapping function
"""

import pickle
import numpy as np
import numpy.linalg as lin
import matplotlib.pyplot as plt
import scipy.special as spe
from assimulo.problem import Implicit_Problem
from assimulo.solvers import IDA

# Initial setup
Pe = 100                                            # Simulation parameters
M  = 32                                             # moments: a_{-M} ... a_1 
N  = 128                                            # Number of quadrature points
Nm = 8                                              # 2*Nm+1 subdivisions
Npt = (2*Nm+1)*N                                    # fine grid
nt  = 1000                                          # Approx number of time points
tol = 1e-7                                          # Solver tolerance
Asy = 'High'                                            # Asymptotic flux ('High'/'Low'/'')
dir = '../data/'                                    # Data directory

# Set end time (traps on a_1)
th = np.sqrt(Pe*np.pi**3/36)                        # High Pe end time
tl = (Pe/2)*(0.5*(1+Pe)-np.euler_gamma-np.log(Pe/4))# Low Pe end time
if Asy == 'High':
    tf = th
elif Asy == 'Low':
    tf = tl
elif Pe > 1:
    tf = th
else:
    tf = tl 
# pick a "nice" interval
ltf = np.log10(tf)
h0 = min([1,2,5,10], key=lambda x:abs(x-10**(ltf%1)))*10**(ltf-ltf%1)/nt
nt = int(tf/h0+1)
tf = nt*h0
# Set name
if Asy == 'Low': M = 0                            # Only 2 moments at low Pe
name = 'Pe='+str(Pe)+'-M='+str(M)
if Asy: name += '-'+Asy
else:
    name += '-N='+str(N)+'-Nm='+str(Nm)
print('Starting run with Pe, M, N, Nm, Asym = ', Pe, M, N, Nm, Asy)
print('Collapse time scale: ', tf)

def K0int(a, b, dp, sgn):                                   # Segment integral
        x  = a + b*dp/2
        ax = np.abs(x)
        sx = np.sign(x) - 1
        return (x/b)*np.exp(sx*ax)*(spe.k0e(ax) + sgn*spe.k1e(ax))           
                
def Kint(cpm, Pea):                                         # Kernel integral
    dp = np.pi/Npt
    p  = np.arange(Npt)*dp + 0.5*dp                         # Fine grid
    cp = np.cos(p)
    sp = np.sin(p)
    a  = Pea*(cpm-cp)
    b  = Pea*sp
    m  = np.argmin(np.abs(a))
    Km = K0int(a[:m],   b[:m],   dp,-1)   \
       - K0int(a[:m],   b[:m],  -dp,-1)                     # n < m
    K0 = K0int(0,       b[m],    dp, 1)   \
       - K0int(0,       b[m],   -dp,-1)                     # n = m
    Kp = K0int(a[m+1:], b[m+1:], dp, 1)   \
       - K0int(a[m+1:], b[m+1:],-dp, 1)                     # n > m
    return (Km.sum() + K0 + Kp.sum())

def flux(a1):
    nM = np.arange(M+2, dtype=int)
    if a1 < 0:  a1 = tol
    if Asy == 'High':
        return  -(2/np.pi)*np.sqrt(a1/(Pe*np.pi))/(nM*nM-1/4)
    if Asy == 'Low':
        sM  = np.zeros(M+2)
        C   = -np.euler_gamma - np.log(Pe*a1/4)
        sM[0] = 2/(Pe*C)
        sM[1] = a1*(1/C - 1) 
        return  sM
    dp = np.pi/N
    Pea= Pe*a1
    p  = np.arange(N)*dp + 0.5*dp                               # set points 0..pi
    cp = np.cos(p)
    K  = np.zeros((N,N))                                        # Kernel matrix
    for m in range(N):
        x  = Pea*(cp[m]-cp)
        ax = np.abs(x)
        sx = np.sign(x) - 1
        K[m] = np.exp(sx*ax)*spe.k0e(ax)*dp
        K[m,m] = 0
        K[m,m] = Kint(cp[m], Pea) - K[m].sum()        
    b  = np.pi*np.ones(N)
    s  = lin.solve(K, b)
    pM = np.outer(nM, p)
    sM = np.dot(np.cos(pM), s)*2*dp    
    return sM/(Pe*np.pi)

def residual(t,y,dy):
    sM  = flux(y[-1])
    res = np.zeros(M+2)
    for m in range(M+2):
        res[m] = sM[m]
        for n in range(-M,2-m):
            res[m] += n*y[n+M]*dy[n+m+M] + (n+m)*y[n+m+M]*dy[n+M]    
    return res 

def state_events(t,y,dy,sw):
     return np.array([y[-1]-tol])

def handle_event(solver, event_info):
    from assimulo.exception import TerminateSimulation
    state_info = event_info[0]
    if state_info[0] != 0:                              # Check if event triggered
        raise TerminateSimulation 

def g():
    t0  = 0.0                                           # Initial conditions
    y0  = np.zeros(M+2)
    y0[-1] = 1
    dy0 = np.zeros(M+2)
    for m in range(M+2):
        dy0[M+1-m] = 8*(Pe*np.pi**3)**(-1/2)/(4*m*m-1)
    dy0[-1] /= 2
    
    model = Implicit_Problem(residual, y0, dy0, t0)     # Create an Assimulo problem
    model.name = 'Conformal mapping coefficients'
    model.state_events = state_events                   #Sets the state events to the problem
    model.handle_event = handle_event                   #Sets the event handling to the problem
  
    sim = IDA(model)
    sim.atol=tol                                        # Create the IDA solver
    sim.rtol=tol
    sim.verbosity=10
    sim.report_continuously=True
    sim.display_progress=True
    sim.print_event_data()

    t,y,dy = sim.simulate(tf, nt) 
    return np.array(t), y

t, y = g()                   # Calculate coefficients of Laurent series A_m(t)
pickle.dump([t, y], open(dir+name+'.pickle', 'wb'))  # Save coefficients

plt.clf()                                          # plot coefficients
for k in range(M-2, M+2):                         #  no more than 5 coeffs
    if k < 0:  continue
    m = k - M
    plt.plot(t/t[-1], y.T[k], label='m = '+str(m))
plt.xlim(0,1)
plt.ylim(0,1)
plt.xlabel('$t$')
plt.ylabel('$a_m(t)$')
plt.legend(loc='best')

